import string
from exceptions import BufferUnderrun


class Buffer:
    buff = b""
    pos = 0

    def __init__(self, data: bytes | bytearray = None):
        if isinstance(data, bytes):
            self.buff = data
        elif isinstance(data, bytearray):
            self.buff = bytes(data)

    def __len__(self):
        return len(self.buff) - self.pos

    def flush(self):
        pass

    def clear(self):
        self.buff = b''
        self.pos = 0

    def copy(self):
        copy_buff = self.__class__(self.buff)
        copy_buff.pos = self.pos
        return copy_buff

    def add(self, data: bytes):
        """
        Add some bytes to the end of the buffer.
        """

        self.buff += data
    
    write = add

    def save(self):
        """
        Saves the buffer contents.
        """

        self.buff = self.buff[self.pos:]
        self.pos = 0

    def restore(self):
        """
        Restores the buffer contents to its state when :meth:`save` was last
        called.
        """

        self.pos = 0

    def discard(self):
        """
        Discards the entire buffer contents.
        """

        self.pos = len(self.buff)

    def read(self, length: int = None):
        """
        Read *length* bytes from the beginning of the buffer, or all bytes if
        *length* is ``None``
        """

        if length is None:
            data = self.buff[self.pos:]
            self.pos = len(self.buff)
        else:
            if self.pos + length > len(self.buff):
                raise BufferUnderrun

            data = self.buff[self.pos:self.pos+length]
            self.pos += length

        return data

    def hexdump(self):
        data = self.buff[self.pos:]
        lines = ['']
        bytes_read = 0
        while len(data) > 0:
            data_line, data = data[:16], data[16:]

            l_hex = []
            l_str = []
            for i, c in enumerate(data_line):
                l_hex.append(f"{c:02x}")
                c_str = data_line[i:i + 1]
                l_str.append(c_str if c_str in string.printable else ".")

            l_hex.extend(['  '] * (16 - len(l_hex)))
            l_hex.insert(8, '')

            lines.append(f"{bytes_read:08x}  {' '.join(l_hex)}  |{''.join(l_str)}|")

            bytes_read += len(data_line)

        return "\n    ".join(lines + [f"{bytes_read:08x}"])
