from protocol.protocolbase import PortocolBase
from buffer import Buffer
from bot import bots, botbase
from consts import RESOURCE_PACK_STATUS, STATES
from typesdata.v754 import Json, VarInt, String, U16, UUID, Bool, F64, F32, I8, Byte, UByte
from router.packetsrouter import PacketHeandler, PacketRouter
import asyncio
from math import log2
from typing import Coroutine, Any, Union
from exceptions import ProtocolError
from mcd.packets import Coder


class ProtocolState(PortocolBase):
    async def response_server(self, byff: Buffer):
        self.logger.info(Json.decode(byff))
        self.bot.close()
    
    def run(self):
        self.on(PacketHeandler('server_info'), self.response_server)
        
        self.loop.create_task(self.asend_packet('ping_start'))
        super(ProtocolState, self).run()


class ProtocolHandshake(PortocolBase):
    next_state = STATES.LOGIN

    def connection_made(self, transport: asyncio.Transport) -> None:
        super(ProtocolHandshake, self).connection_made(transport)
        self.send_packet('set_protocol',
                        VarInt.encode(self.bot.mcd.protocol_version),
                        String.encode(self.bot.ip),
                        U16.encode(self.bot.port),
                        VarInt.encode(self.next_state)
                        )
        self.bot.switch_protocol(self.next_state)


class ProtocolAuth(PortocolBase):
    bot: 'bots.BotAuth'

    async def set_compression(self, byff: Buffer):
        value = VarInt.decode(byff)
        self.bot.compression_threshold = value
        self.logger.debug(f'Set compression to {value}')

    async def disconect(self, byff: Buffer):
        self.logger.info(f'Disconect in mode {self.state} {Json.decode(byff)}')

    def success(self, byff: Buffer):
        self.bot.uuid = UUID.decode(byff)
        self.logger.info('Client login')

    def success_and_switch_right(self, buff: Buffer):
        self.bot.switch_right()
        self.success(buff)

    def run(self):
        super(ProtocolAuth, self).run()
        self.on('disconnect', self.disconect)

        if not self.bot.profile.online:
            self.once('compress', self.set_compression)
            
            self.once('success', self.success_and_switch_right)
            self.loop.create_task(self.asend_packet('login_start', String.encode(self.bot.profile.display_name)))
        
        else:
            raise NotImplementedError('online mode not implemented')


class ProtocolSpawn(PortocolBase):
    bot: 'bots.BotSpawn'
    
    update_po: Union[asyncio.Task, None] = None
    update_pl: Union[asyncio.Task, None] = None

    async def update_player_inc(self):
        while not self.bot.is_close:
            await asyncio.sleep(1 / 20)
            await self.asend_packet('client_command', Bool.encode(True))

    async def keep_alive(self, buff: Buffer):
            self.send_packet('keep_alive', buff.read())

    async def set_position_look(self, buff: Buffer):
        position_look = (F64.decode(buff), F64.decode(buff), F64.decode(buff), F32.decode(buff), F32.decode(buff))
        flag_b = I8.decode(buff)
        
        if not not flag_b:
            flag = int(log2(flag_b))
            xyz = self.bot.position_look[flag]

            self.bot.position_look.__init__(position_look)
            self.bot.position_look[flag] += xyz
        else:
            self.bot.position_look.__init__(position_look)

        await self.asend_packet('teleport_confirm', buff.read())

    async def update_position_and_look(self):
        while not self.bot.is_close:
            await asyncio.sleep(1)
            await self.asend_packet(
                "position_look",
                F64.encode(self.bot.position_look[0]),
                F64.encode(self.bot.position_look[1] - 1.62),
                F64.encode(self.bot.position_look[2]),
                F32.encode(self.bot.position_look[3]),
                F32.encode(self.bot.position_look[4]),
                Bool.encode(True))

    def done_callback(self, task: asyncio.Task):
        setattr(self, task.get_name(), None)

    def create_set_task(self, name_attr: str, coro: Coroutine[Any, Any, None]):
        attr = getattr(self, name_attr, None)
        if isinstance(attr, asyncio.Task) and not attr.done():
            return 
        task = self.loop.create_task(coro, name=name_attr)
        setattr(self, name_attr, task)
        task.add_done_callback(self.done_callback)

    async def first_set_position_look(self, byff: Buffer):
        await self.set_position_look(byff)
        self.on('position', self.set_position_look)
        self.create_set_task('update_po', self.update_position_and_look())

    async def disconect(self, byff: Buffer):
        self.logger.info(f'Disconect in mode {self.state} {Json.decode(byff)}')

    def switch(self):
        super(ProtocolSpawn, self).switch()
        if not self.update_po is None: self.update_po.cancel()
        if not self.update_pl is None: self.update_pl.cancel()

    def run(self):
        super(ProtocolSpawn, self).run()
        self.bot.position_look = [0.0, 0.0, 0.0, 0.0, 0.0]
        self.once('kick_disconnect', self.disconect)
        self.on('keep_alive', self.keep_alive)
        self.create_set_task('update_pl', self.update_player_inc())

        self.once('position', self.first_set_position_look)
        
        self.loop.create_task(self.asend_packet('resource_pack_receive', VarInt.encode(self.bot.settings.resource_pack_status)))
        
        clientsettings = self.bot.settings.clientsettings
        
        self.loop.create_task(self.asend_packet('settings', String.encode(clientsettings.locale),
                                                Byte.encode(clientsettings.view_distance),
                                                VarInt.encode(clientsettings.chat_mode),
                                                Bool.encode(clientsettings.chat_colors),
                                                UByte.encode(clientsettings.displayed_skin_parts),
                                                VarInt.encode(clientsettings.main_hand)))
        
        self.loop.create_task(self.asend_packet('custom_payload', self.bot.settings.custom_payload))



