import abc 
from typing import Callable, Union, Coroutine, Any
from asyncio import AbstractEventLoop, iscoroutinefunction, Task
from exceptions import BaseRouterExceprion, BaseCallbackExceprion

TYPES_EVENT = Union[str, bytes, int, Any]


class Callback:
    run: Callable[[Any], Task | None]
    _origin_func: Callable
    once: bool
    
    def __init__(self, loop: AbstractEventLoop, func: Union[Callable[[Any], Any], Callable[[Any], Coroutine[Any, Any, Any]]], once: bool):
        self.once = once
        self._origin_func = func
        
        if not isinstance(func, Callable):
            raise BaseCallbackExceprion(f'func is not Callable')
        elif iscoroutinefunction(func):
            self.run = lambda buffer: loop.create_task(func(buffer), name=func.__name__)
        else:
            self.run = func


class CallbackUnheandler(Callback):
    def __init__(self, loop: AbstractEventLoop, func: Union[Callable[[Any], Any], Callable[[Any], Coroutine[Any, Any, Any]]]):
        self.once = False
        self._origin_func = func

        if not isinstance(func, Callable):
            raise BaseCallbackExceprion(f'func is not Callable')
        elif iscoroutinefunction(func):
            self.run = lambda event_id: lambda buffer: loop.create_task(func(event_id, buffer), name=func.__name__)
        else:
            self.run = func


class BaseStorage(abc.ABC):
    _unheandler: CallbackUnheandler | None = None

    @abc.abstractmethod
    def __init__(self):
        ...
    
    def set(self, event_id: TYPES_EVENT, callback: Callback):
        self._storage[event_id].add(callback)
    
    def _get(self, event_id: TYPES_EVENT) -> list[Callback]:
        callbacks = self._storage[event_id]
        
        if callbacks:
            callbacks_ret = list(callbacks.copy())
            for cl in callbacks_ret:
                if cl.once:
                    callbacks.remove(cl)
        else:
            callbacks_ret = []
        return callbacks_ret
    
    def _get_unh(self, event_id: TYPES_EVENT) -> list[Callback]:
        callbacks = self._storage[event_id]
        
        if callbacks:
            callbacks_ret = list(callbacks.copy())
            for cl in callbacks_ret:
                if cl.once:
                    callbacks.remove(cl)
        else:

            callbacks_ret = [self._unheandler.run(event_id)]
        return callbacks_ret

    get = _get

    def clear(self):
        for callbacks in self._storage:
            callbacks.clear()

    @property
    def unheandler(self):
        return self._unheandler
    
    @unheandler.setter
    def unheandler(self, func):
        if func is None:
            self.get = self._get
        elif isinstance(func, CallbackUnheandler):
            self.get = self._get_unh
            self.unheandler = func
        else:
            raise BaseRouterExceprion(f'<unheandler> is already type CallbackUnheandler')

    def remove(self, event_id: TYPES_EVENT, fn):
        callbacks = self._storage[event_id]
        
        for cl in callbacks.copy():
            if cl._origin_func is fn:
                callbacks.remove(fn)


class BaseRouter:
    def __init__(self,
                 loop: AbstractEventLoop,
                 storage: BaseStorage,
                 unheandler: Callable = None
                 ):
        self.loop = loop
        self.storage = storage
        if unheandler is not None: storage.unheandler = CallbackUnheandler(loop, unheandler)
        self._tasks_set: set[Task] = set()

    def clear(self):
        for task in self._tasks_set.copy():
            task.cancel()
        self.storage.clear()

    def on(self, event_id: TYPES_EVENT, callback: Callable):
        self.storage.set(event_id, Callback(self.loop, callback, False))
        return callback

    def once(self, event_id: TYPES_EVENT, callback):
        self.storage.set(event_id, Callback(self.loop, callback, True))
        return callback
        
    def rout(self, event_id: TYPES_EVENT, data):
        callbacks = self.storage.get(event_id)
        if callbacks:
            for callback in callbacks:
                callback_task = callback.run(data)
                if isinstance(callback_task, Task):
                    self._tasks_set.add(callback_task)
                    def _done(task: Task):
                        self._tasks_set.remove(task)
                    
                    callback_task.add_done_callback(_done)


class BaseHeandler:
    target_router = BaseRouter

    def __init__(self, event: TYPES_EVENT):
        self.value = event

    def __str__(self):
        return f'{self.__class__.__name__}<{self.value}>'
    
    def __repr__(self):
        return f'{self.__class__.__name__}<{self.value.__repr__()}>'







